package com.tidecloud.dataacceptance.service.impl;

import com.accept.client.DeviceMsgClient;
import com.accept.model.DeviceMsgDto;
import com.smartsanitation.common.util.StringUtil;
import com.tidecloud.dataacceptance.codec.HeaderTailDelimiterFrameDecoder;
import com.tidecloud.dataacceptance.common.*;
import com.tidecloud.dataacceptance.entity.YiTongGPSDevice;
import com.tidecloud.dataacceptance.service.HexBinaryAcceptanceHandlerAdapter;
import com.tidecloud.dataacceptance.util.FileUtils;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.xml.bind.DatatypeConverter;
import java.util.Calendar;
import java.util.Date;

/**
 * @author jhw
 * GK309 卡牌数据 康凯斯
 */
@Component
@ChannelHandler.Sharable
@Scope("prototype")
public class GK309GpsServerHandler extends HexBinaryAcceptanceHandlerAdapter {

    /**
     * 协议信息数据
     * [ 起始位|包长度|协议号|信息内容|信息序列号|错误校验|停止位 ]
     */
    private static final Logger logger = LoggerFactory.getLogger(GK309GpsServerHandler.class);

    // 起始位
    private static final Integer START_BITS = 2;// 起始位长度
    private static final Byte START_BIT = 0x78; //单字节长度数据包
    private static final Byte START_BIT2 = 0X79;// 双字节长度数据包

//    private static final int START_BIT = 120; //单字节长度数据包
//    private static final int START_BIT2 = 121;// 双字节长度数据包

    // 协议号
    private static final byte LOGIN_MSG = 0x01;// 登录 (必须回复)
    private static final byte STATUS_MSG = 0x13;// 心跳，状态信息包  (必须回复)
    private static final byte LOCATION_MSG = 0x10;// 一般GPS 位置信息上传
    private static final byte CORRECT_TIME_MSG = 0x1F;// 校验时间
    private static final byte SIGN_IN_MSG = (byte)0xB0;// 打卡

    private static final byte[] END_BITS = new byte[]{0X0D, 0x0A};

    private static final Integer DATA_SIZE = 6;

    private static final byte MARK = 0B00111100;

    private static final byte SIGN_IN = 0B00000100; //上班打卡
    private static final byte SIGN_OUT = 0B00001000; // 下班打卡

//    @Autowired
    private DeviceMsgClient deviceMsgClient = new DeviceMsgClient() {
    @Override
    public void acceptDeviceMsg(DeviceMsgDto deviceMsgDto) {
        logger.info(StringUtil.convert2String(deviceMsgDto));

    }

    @Override
    public void acceptDeviceMsgParam(String s, String s1, Integer integer, String s2, Long aLong) {
        logger.info("111");

    }
};

    @Override
    protected void handle(ByteBuf in, Channel channel) throws Exception {
        if (in.isReadable()) {
            in.markReaderIndex();
            try {
                byte b = 0;
                int index = 0;
                while (index < START_BITS) {
                    b = in.readByte();
                    if (START_BIT2 != b && START_BIT != b) {
                        channel.close();
                    }
                    index++;
                }
                /*  单字节长度或者双字节长度 */
                int length = 0;
                if (START_BIT == b) {
                    length = in.readByte() & 0xff; // 一个字节
                } else {
                    length = in.readShort() & 0xffff;// 两个字节
                }
                handle(in, length, channel);
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
        }
    }

    private void handle(ByteBuf in, int length, Channel channel) throws Exception {
        if (in.isReadable()) {
            byte msgType = in.readByte();// 协议号 1个字节
            String deviceId = channelDeviceMap.get(channel);
            if (deviceId != null) {
                MDC.put(MDC_DEVICEID, deviceId);
            } else {
                if (LOGIN_MSG != msgType) {
                    channel.close();
                }
            }
            if (LOGIN_MSG == msgType) {
                resolveLoginMSG(in, channel);
            } else if (STATUS_MSG == msgType) {
                reply(channel, STATUS_MSG, "心跳包");
                byte[] bytes = byteMerger(getOriginalData(in, deviceId), DateUtil.formatDate2String(new Date()).getBytes());
                sendMsg2Kafka(bytes, deviceId, channel);// 心跳数据对接
                deviceMsgClient.acceptDeviceMsgParam("kangkaisi",
                        deviceId,1,
                        DatatypeConverter.printHexBinary(bytes), System.currentTimeMillis());
            } else if (LOCATION_MSG == msgType) {
                logger.info("GPS 定位包");
                // resolveLocationMSG(in,deviceId);
                byte[] bytes = byteMerger(getOriginalData(in, deviceId), DateUtil.formatDate2String(new Date()).getBytes());
                sendMsg2Kafka(bytes, deviceId, channel);
                deviceMsgClient.acceptDeviceMsgParam("kangkaisi",
                        deviceId,2,
                        DatatypeConverter.printHexBinary(bytes), System.currentTimeMillis());
            } else if (CORRECT_TIME_MSG == msgType) {
                // 1位长度位和6位时间位,后面才是序列号
                in.skipBytes(7);
                short x = in.readShort();
                byte[] ret = short2Bytes(x);

//                reply(channel, CORRECT_TIME_MSG, "校时包");
                replyForTime(channel, ret);
            } else if (SIGN_IN_MSG == msgType) {
                byte[] dataBytes = new byte[6];
                in.readBytes(dataBytes);

                in.skipBytes(1);
                short terminalReservedWord = in.readShort();
                byte[] reservedWorkBytes = short2Bytes(terminalReservedWord);


                // 根据文档查看上下班标示
                ByteBuf byteBuf = in.readBytes(13);
                byteBuf.skipBytes(12);
                byte b = byteBuf.readByte();
                // 2-5位代表是上班还是下班
                byte i = (byte) (b & MARK);

//                类型: 1(上班), 2(下班)
                Byte signResponse = null;
                if (i == SIGN_IN) {
                    signResponse = 0X01;
                } else {
                    signResponse = 0x02;
                }


                // 减去前面跳过的第一位和前面跳过的7位，再减去最后的四位（序列号本身的2位+校验位2位）
                in.skipBytes(length - 8 - 2 - 2 - 15);
                short serNum = in.readShort();
                byte[] serNumBytes = short2Bytes(serNum);

                // 状态： 0(失败), 1(成功)
                byte[] signResult = new byte[]{0x01, signResponse};
                // 0X0F是长度，固定长度
                byte[] contentsOfPackage = mergeByteArray(new byte[]{0X0F, SIGN_IN_MSG},
                        dataBytes, signResult, reservedWorkBytes, serNumBytes);
                writeToClient(channel, "打卡包", contentsOfPackage);

            } else  {
                logger.info("client send data without handle type ...");
            }
        }

    }

    private byte[] short2Bytes(short x) {
        byte[]  ret = new byte[2];
        ret[0] = (byte)(x & 0xff);
        ret[1] = (byte)((x >> 8) & 0xff);
        return ret;
    }

    /**
     * 登录管理
     *
     * @param in
     * @param channel
     */
    private void resolveLoginMSG(ByteBuf in, Channel channel) {
        /**  终端ID|类型识别码|扩展位*/
        byte[] deviceIdBytes = new byte[8];// 终端ID [8位]
        in.readBytes(deviceIdBytes);
        String deviceId = DatatypeConverter.printHexBinary(deviceIdBytes);
        String deviceIdInMap = channelDeviceMap.get(channel);
        MDC.put(MDC_DEVICEID, deviceId);
        if (!deviceId.equals(deviceIdInMap)) {
            manageChannel(channel, deviceId);
        }
        reply(channel, LOGIN_MSG, "登录包");
    }

    public static byte[] byteMerger(byte[] byte_1, byte[] byte_2) {
        byte[] byte_3 = new byte[byte_1.length + byte_2.length];
        System.arraycopy(byte_1, 0, byte_3, 0, byte_1.length);
        System.arraycopy(byte_2, 0, byte_3, byte_1.length, byte_2.length);
        return byte_3;
    }


    /**
     * 回复包
     *
     * @param channel
     * @param msgType
     */
    private void reply(Channel channel, byte msgType, String msg) {

        /**
         * 起始位[]长度[] 协议号[]	序列号[]CRC校验[]停止位[2]
         */
        ByteBuf buffer = Unpooled.buffer();
        byte[] crcBytes = new byte[]{0x05, msgType, 0x00, 0x05}; // CRC校验数据
        int doCrc = CRCUtil.do_crc(65535, crcBytes);//CRC-ITU
        byte[] intToByte = NumUtil.intToByte(doCrc, 2);
        byte[] bytes = new byte[]{0x78, 0x78, 0x05, msgType, 0x00, 0x05, intToByte[0], intToByte[1], 0x0D, 0x0A};
        buffer.writeBytes(bytes);
        ChannelFuture channelFuture = channel.write(buffer);
        channelFuture.addListener(future -> {
            String deviceId = channelDeviceMap.get(channel);
            logger.info("server reply [{}] to client device [{}] success:[{}] ", msg, deviceId, DatatypeConverter.printHexBinary(bytes));
        });
    }

    private byte[] mergeByteArray(byte[]... bytesArray) {
        byte[] result = ArrayUtils.EMPTY_BYTE_ARRAY;
        for (byte[] bytes : bytesArray) {
            result = ArrayUtils.addAll(result, bytes);
        }
        return result;
    }

    /**
     * 校时回复
     * @param channel
     */
    private void replyForTime(Channel channel, byte[] serialNumber) {
        byte msgType = CORRECT_TIME_MSG;
        String msg = "校验时间";

        /**
         * 起始位[]长度[] 协议号[]	序列号[]CRC校验[]停止位[2]
         */
        long utcSeconds = Calendar.getInstance().getTimeInMillis() / 1000L;

        // 长度(固定0B)-协议号(固定）-utc时间秒数-预留位(2个0x00)-序列号
        byte[] utcBytes = BitOperator.longToBytes(utcSeconds, 4);
        byte[] seriNumbytes = mergeByteArray(new byte[]{0x00, 0x00}, serialNumber);
        byte[] contentsOfPackage = mergeByteArray(new byte[]{0x0B, msgType}, utcBytes, seriNumbytes);

        writeToClient(channel, msg, contentsOfPackage);
    }

    private void writeToClient(Channel channel, String msg, byte[] contentsOfPackage) {
        ByteBuf buffer = Unpooled.buffer();
        byte[] bytesToReply = wrapeContents(contentsOfPackage);
        buffer.writeBytes(bytesToReply);
        ChannelFuture channelFuture = channel.write(buffer);
        channelFuture.addListener(future -> {
            String deviceId = channelDeviceMap.get(channel);
            logger.info("server reply [{}] to client device [{}] success:[{}] ", msg, deviceId, DatatypeConverter.printHexBinary(bytesToReply));
        });
    }

    private byte[] wrapeContents(byte[] contentsOfPackage) {
        int doCrc = CRCUtil.do_crc(65535, contentsOfPackage);//CRC-ITU
        byte[] crcBytes = NumUtil.intToByte(doCrc, 2);

        return mergeByteArray(new byte[]{START_BIT, START_BIT}, contentsOfPackage, crcBytes, END_BITS);
    }

    /**
     * 原始数据重新组装
     *
     * @param in
     * @param deviceId
     * @return
     */
    private byte[] getOriginalData(ByteBuf in, String deviceId) {
        if (StringUtils.isNotBlank(deviceId)) {
            in.resetReaderIndex();
            byte[] deviceArr = deviceId.getBytes();
            int length = in.readableBytes();
            byte[] dataByteArray = new byte[length + deviceArr.length];
            in.readBytes(dataByteArray, 0, in.readableBytes());
            System.arraycopy(deviceArr, 0, dataByteArray, length, deviceArr.length);
            return dataByteArray;
        } else
            return null;
    }

    private void resolveLocationMSG(ByteBuf in, String deviceId) throws Exception {
        YiTongGPSDevice yiTongGPSDevice = new YiTongGPSDevice();
        StringBuffer dateTimeStrBuf = new StringBuffer();
        int indexOfDateTime = 0;
        while (indexOfDateTime < DATA_SIZE) {
            byte b = in.readByte();
            dateTimeStrBuf.append(NumUtil.byte2String(b));
            indexOfDateTime++;
        }
        // 日期
        yiTongGPSDevice.setDate(dateTimeStrBuf.toString());
        // gps信息卫星数
        yiTongGPSDevice.setGpsCount(in.readByte());
        // 维度
        yiTongGPSDevice.setLat(in.readInt());
        // 经度
        yiTongGPSDevice.setLng(in.readInt());
        // 速度
        yiTongGPSDevice.setSpeedbyte(in.readByte());
        // 航向
        yiTongGPSDevice.setCourseStatus(in.readShort());
        // 国家代号
        yiTongGPSDevice.setMcc(in.readShort());
        // 移动网号码
        yiTongGPSDevice.setMnc(in.readByte());
        // 位置区码
        yiTongGPSDevice.setLac(in.readShort());
        // 移动基站Cell Tower ID
        yiTongGPSDevice.setCellId(in.readMedium());
        yiTongGPSDevice.setAcc(in.readByte());
        // 数据上报模式 0x00:定时上报,0x01:定距上报,0x02:拐点上传,0x03:ACC状态改变上传,0X08:开机上报位置信息
        yiTongGPSDevice.setReportModel(in.readByte());
        // 0x01:实时 0x00:补传
        yiTongGPSDevice.setIsmendMsg(in.readByte());
        double mileage = NumUtil.toFixed2Place((double) in.readInt() / 1000);
        // 里程设备默认是关闭的，需要指令，设备端才发送
        yiTongGPSDevice.setDeviceId(deviceId);
        yiTongGPSDevice.setMileage(mileage);
        yiTongGPSDevice.setDataType(1);
        // 写文件操作
        String deviceStr = YiTongGPSDevice.buildDeviceStr(yiTongGPSDevice);
        FileUtils.dataStorage(deviceStr.getBytes(), dataPath, prefixName);
    }

    /**
     * 按照开始，结束标记位读取数据
     */
    @Override
    public void startAcceptor() {
        EventLoopGroup bossGroup = new NioEventLoopGroup();
        EventLoopGroup workerGroup = new NioEventLoopGroup();
        byte[] tailSplitBytes = new byte[]{0x0D, 0x0A};
        byte[] headerSplitBytes = new byte[]{0x78, 0x78};
        byte[] headerSplitBytes1 = new byte[]{0x79, 0x79};
        try {
            ServerBootstrap b = new ServerBootstrap();
            b.group(bossGroup, workerGroup).channel(NioServerSocketChannel.class)
                    .childHandler(new ChannelInitializer<SocketChannel>() {
                        @Override
                        protected void initChannel(SocketChannel ch) throws Exception {
                            ch.pipeline().addLast(new HeaderTailDelimiterFrameDecoder(65535, false,
                                    Unpooled.copiedBuffer(tailSplitBytes), Unpooled.copiedBuffer(headerSplitBytes),
                                    Unpooled.copiedBuffer(headerSplitBytes1)));
                            ch.pipeline().addLast(GK309GpsServerHandler.this);
                        }
                    }).option(ChannelOption.SO_BACKLOG, 128).childOption(ChannelOption.SO_KEEPALIVE, true);
            ChannelFuture f = b.bind(port).sync();
            f.channel().closeFuture().sync();
            logger.info("start accept service for {}, bind address {}:{}", this.getPrefixName(), this.getIp(),
                    this.getPort());
        } catch (InterruptedException e) {
            logger.error(e.getMessage());
        } finally {
            cleanRedisLinkData();
            workerGroup.shutdownGracefully();
            bossGroup.shutdownGracefully();
        }
    }

}
